//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                          //
//                                           Calculator  TDS3T                                              //
//                                                                                                          //
//                                          Copyright © 2023 by                                             //
//                                        Zumwalt Properties, LLC.                                          //
//                                          All Rights Reserved.                                            //
//                                                                                                          //
//          This software is for NON COMMERCIAL USE ONLY in accordance with the licensing agreement         //
//                                     as published on this website.                                        //
//                                                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Included files.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define   TOUCH_MODULES_CST_SELF
#include  <TouchLib.h>

#include  <TFT_eSPI.h>
#include  <Wire.h>
#include  "Greg on pole small.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Global constants.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Buttons.

#define BUTTON_WIDTH            42
#define BUTTON_HEIGHT           42

// Debug.

#define DEBUG                   false

// Display.

#define CLEAR_KEY_PRESSED_DELAY 40
#define DISPLAY_BRIGHTNESS_MAX  254
#define DISPLAY_BRIGHTNESS_MIN  0
#define DISPLAY_HEIGHT          320
#define DISPLAY_WIDTH           170
#define DISPLAY_ORIENTATION     0
#define FONT_SIZE               2
#define PIN_BRIGHTNESS_UP       14
#define PIN_BRIGHTNESS_DOWN     0
#define PIN_POWER_ON            15
#define PIN_LCD_BL              38
#define READOUT_CHARACTER_WIDTH 14

// Touch.

#define PIN_TOUCH_INT           16
#define PIN_TOUCH_RES           21
#define TOUCH_MAX_X             310
#define TOUCH_MAX_Y             160

// Wire.

#define PIN_IIC_SCL             17
#define PIN_IIC_SDA             18

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Global variables.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Display.

bool                        bClearKeyPressed = false;                               // Clear key pressed.
int                         nClearKeyPressedCount = CLEAR_KEY_PRESSED_DELAY;        // Clear key pressed counter.
TFT_eSPI                    displayTFT = TFT_eSPI();                                // Tft display.
TFT_eSprite                 spriteBackground = TFT_eSprite(& displayTFT);           // Background sprite.

// Interrupts.

hw_timer_t *                timerInterruptService = NULL;                           // Interrupt service timer hook.
volatile SemaphoreHandle_t  timerLoopSemaphore;                                     // Main loop semaphore from interrupt service.

// Keypad.
  
  // Button background colors starting with the top row.
  
  int  chColors[] =         {TFT_DARKGREY, TFT_DARKGREY, TFT_DARKGREY, TFT_ORANGE,
                             TFT_DARKGREY, TFT_DARKGREY, TFT_DARKGREY, TFT_ORANGE,
                             TFT_DARKGREY, TFT_DARKGREY, TFT_DARKGREY, TFT_ORANGE,
                             TFT_DARKGREY, TFT_DARKGREY, TFT_ORANGE, TFT_ORANGE
                            };
  
  // Button symbols starting with the top row.
  
  char chSymbols[] =       {'7', '8', '9', '*',
                            '4', '5', '6', '/',
                            '1', '2', '3', '+',
                            '0', '.', '=', '-'
                           };
  

// LCD brightness.

int                         lcdBacklightBrightness = DISPLAY_BRIGHTNESS_MAX;        // Display brightness.

// Math.

bool                        bEqual = false;                                         // Equal sign pressed.
bool                        bError = false;                                         // Error flag (e.g. division by zero).
char                        chReadout[81] = {'0', '.', 0};                          // Calculator readout buffer.
int                         nReadoutPosition = 0;                                   // Calculator readout character position.
bool                        bOperator = false;                                      // Operator pressed.
double                      dOperands[2];                                           // Operands.
int                         nOperands = 0;                                          // Operand count.
char                        chOperator;                                             // Operator.
int                         nOperatorStackPosition = 0;                             // Operator stack position.

// Touch.

TouchLib                    tlTouch(Wire,                                           // Touch library.
                                    PIN_IIC_SDA,                                    // Data.  
                                    PIN_IIC_SCL,                                    // Clock.
                                    CTS820_SLAVE_ADDRESS,                           // Slave address.
                                    PIN_TOUCH_RES);                                 // Reset pin.
TP_Point                    tpTouch;                                                // Touch point coordinates.
bool                        bTouchpadPressed = false;                               // Touch press detected.
int                         nTouchN1 = 0;                                           // Previous touch state.
int                         nTouchX = 0;                                            // Display compensated touch X.
int                         nTouchY = 0;                                            // Display compensated touch Y.

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// InterruptService().
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// InterruptService
// Called by timer interrupt services for updating of main loop semaphore.
// Entry   : nothing
// Returns : nothing
// Notes   :
//
//  1) Designed for a 1000hz (1ms) interrupt.
//

void IRAM_ATTR InterruptService()
{
    // Local constants.
    
    #define   MAIN_DELAY          100

    // Local variables.
      
    static  unsigned long   nMainDelay = MAIN_DELAY - 1;

    // Check nMainDelay.
    
    if(nMainDelay)
    {
      // Not yet time to update.
      
      nMainDelay --;
    }
    else
    {
      // Time to update, reset delay.
      
      nMainDelay = MAIN_DELAY - 1;
  
      // Execute a main loop pass.
        
      xSemaphoreGiveFromISR(timerLoopSemaphore, NULL);
    }

  // End.
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// setup().
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void setup() 
{
  // Begin serial.

  #if DEBUG
  Serial.begin(115200);
  #endif

  // Enable battery operation.
  
  pinMode(PIN_POWER_ON, OUTPUT);
  digitalWrite(PIN_POWER_ON, HIGH);

  // Enable sleep button input.

  pinMode(0, INPUT_PULLUP);
  
  // Reset touch.
  
  pinMode(PIN_TOUCH_RES, OUTPUT);
  digitalWrite(PIN_TOUCH_RES, LOW);
  delay(500);
  digitalWrite(PIN_TOUCH_RES, HIGH);

  // Initialize the tft display.
  
  displayTFT.begin();
  displayTFT.setRotation(DISPLAY_ORIENTATION);
  displayTFT.setSwapBytes(true);
  displayTFT.setTextSize(FONT_SIZE);
  displayTFT.setTextColor(TFT_WHITE, TFT_BLACK);
  displayTFT.fillScreen(TFT_BLACK);

  // Initialize lcd.

  pinMode(PIN_BRIGHTNESS_UP, INPUT_PULLUP);
  pinMode(PIN_BRIGHTNESS_DOWN, INPUT_PULLUP);
  ledcSetup(0, 10000, 8);
  ledcAttachPin(PIN_LCD_BL, 0);
  ledcWrite(0, lcdBacklightBrightness);

  // Create a background sprite.
  
  spriteBackground.createSprite(DISPLAY_WIDTH, DISPLAY_HEIGHT);
  spriteBackground.setSwapBytes(false);
  spriteBackground.setTextColor(TFT_BLACK, TFT_DARKGREY);
  spriteBackground.setTextSize(FONT_SIZE);

  // Begin wire service to the touch hardware.
  
  Wire.begin(PIN_IIC_SDA, PIN_IIC_SCL);

  // Initialize the touch hardware.
  
  if(!tlTouch.init())
  {
    #if DEBUG
    Serial.println("Touch IC not found");
    #endif
  }
  tlTouch.setRotation(DISPLAY_ORIENTATION);
  
 // Initialize interrupt service.

    // Create timerInterruptService to main loop semaphore.
    
    timerLoopSemaphore = xSemaphoreCreateBinary();
    
    // Use timer 0 (4 available from 0 through 3), with a prescale of 80.
    
    timerInterruptService = timerBegin(0, 80, true);
    
    // Attach InterruptService() to timer.
    
    timerAttachInterrupt(timerInterruptService, & InterruptService, true);
    
    // Set alarm to call onTimer function every millisecond (value in microseconds)
    // with repeat (third parameter).
    
    timerAlarmWrite(timerInterruptService, 1000, true);
    
    // Start the alarm.
    
    timerAlarmEnable(timerInterruptService);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// loop().
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void loop(void)
{
  // Initialize the display.

  DrawDisplay();

  // Then loop forever.
  
  while(1)
  {
    // Wait for InterruptService() semaphore.
    
    if(xSemaphoreTake(timerLoopSemaphore, portMAX_DELAY) == pdTRUE)
    {
      // Update touch status.
  
      if(UpdateTouchStatus())
      {
        if(bTouchpadPressed)
        {
          // A touchpad press occured.

          DrawKeypad();
          
          // Process key presses.
          
          ProcessKeyPresses(ScanKeypad());

          // Reset touch pressed.
  
          bTouchpadPressed = false;
        }
        else
        {
          // A touch release occured, draw the display to unhighlight the pressed key.
  
          DrawDisplay();
        }
      }

      // Check for sleep.

      if(bClearKeyPressed)
      {
        // Clear key is pressed, update the clear key pressed counter
        
        nClearKeyPressedCount --;

        // Display the countdown.

        if(nClearKeyPressedCount < (CLEAR_KEY_PRESSED_DELAY - 10))
        {
          spriteBackground.setTextSize(8);
          spriteBackground.setTextColor(TFT_BLACK, TFT_WHITE);
          char  chBuffer[81];
          sprintf(chBuffer, "%1d", ((nClearKeyPressedCount + 10) / 10));
          spriteBackground.drawString(String(chBuffer), 
                                      (DISPLAY_WIDTH / 2) - (spriteBackground.textWidth(chBuffer) / 2), 
                                      (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4) - (displayTFT.fontHeight() + 6)) / 2 - (spriteBackground.fontHeight() / 2));
          spriteBackground.setTextSize(FONT_SIZE);
        }
        
        // Check for time to sleep.
        
        if(nClearKeyPressedCount == 0)
        {
          // Reset the clear key parameters.
          
          bClearKeyPressed = false;
          nClearKeyPressedCount = CLEAR_KEY_PRESSED_DELAY;

          // Erase the display.
          
          displayTFT.fillScreen(TFT_BLACK);

          // LCD off.

          digitalWrite(PIN_POWER_ON, LOW);

          // Delay.

          delay(5000);
          
          // Enable touch sleep.
          
          tlTouch.enableSleep();

          // Sleep.
          
          gpio_hold_en((gpio_num_t)PIN_TOUCH_RES);
          esp_sleep_enable_ext0_wakeup((gpio_num_t)PIN_TOUCH_INT, 0);
          esp_deep_sleep_start();
        }
      }
      else
      {
        // Clear key is not pressed.
        
        nClearKeyPressedCount = CLEAR_KEY_PRESSED_DELAY;
      }

      // Adjust display brightness.
      
      if(digitalRead(PIN_BRIGHTNESS_UP) == 0)
      {
        // Increase.
        
        lcdBacklightBrightness += 8;
        lcdBacklightBrightness = (lcdBacklightBrightness <= DISPLAY_BRIGHTNESS_MAX) ? lcdBacklightBrightness : DISPLAY_BRIGHTNESS_MAX;
        ledcSetup(0, 10000, 8);
        ledcAttachPin(38, 0);
        ledcWrite(0, lcdBacklightBrightness);
      }
      else if(digitalRead(PIN_BRIGHTNESS_DOWN) == 0)
      {
        // Decrease.
        
        lcdBacklightBrightness -= 8;
        lcdBacklightBrightness = (lcdBacklightBrightness >= DISPLAY_BRIGHTNESS_MIN) ? lcdBacklightBrightness : DISPLAY_BRIGHTNESS_MIN;
        ledcSetup(0, 10000, 8);
        ledcAttachPin(38, 0);
        ledcWrite(0, lcdBacklightBrightness);
      }

      // Update the tft.
  
      spriteBackground.pushSprite(0, 0);
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Utilities.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// DrawDisplay
// Update the display.
// Entry   : nothing
// Returns : nothing
//

void  DrawDisplay()
{
  // Local variables.

  static  bool    bError = false;
                                                        
  // Erase the background sprite.
  
  spriteBackground.setSwapBytes(true);
  spriteBackground.pushImage(-60, -40, 320, 170, Greg_on_pole_small);
  spriteBackground.setSwapBytes(false);

  // Draw the keypad onto the background sprite.

  DrawKeypad();

  // Draw the readout onto the background sprite.

  DrawReadout();
}

//
// DrawKeypad
// Draw the keypad.
// Entry   : nothing
// Returns : nothing
//

void  DrawKeypad()
{
  // Compute the touched kepad button, or 'C' if the touch occured off the keypad.
  
  int     nButtonTouched = (tpTouch.y > (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4))) 
                            ? (tpTouch.x / BUTTON_WIDTH) + ((tpTouch.y - (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4))) / BUTTON_HEIGHT * 4)
                            : 'C';

  // Draw the keyboard starting by scanning the four rows.
  
  for(int nRow = 0; nRow < 4; nRow ++)
  {
    // And then the four columns on the current row.
    
    for(int nColumn = 0; nColumn < 4; nColumn ++)
    {
      // Draw this row, column.
      
      int   nButton = (nRow * 4) + nColumn;
      int   nX = nColumn * BUTTON_WIDTH;
      int   nY = (nRow * BUTTON_HEIGHT) + (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4));

      // To highlight this button or not.
      
      if((nButton == nButtonTouched) && (bTouchpadPressed))
      {
        // Highlight this button.
        
        spriteBackground.fillRoundRect(nX, nY, BUTTON_WIDTH - 2, BUTTON_HEIGHT - 2, 4, TFT_WHITE);
        spriteBackground.setTextColor(TFT_BLACK, TFT_WHITE);
        spriteBackground.drawChar(chSymbols[(nRow * 4) + nColumn], 
                                  nX + (BUTTON_WIDTH / 2) - (spriteBackground.textWidth(String(chSymbols[nButton])) / 2), 
                                  nY + (BUTTON_HEIGHT / 2) - (spriteBackground.fontHeight() / 2));
      }
      else
      {
        // Do not highlight this button.
        
        spriteBackground.fillRoundRect(nX, nY, BUTTON_WIDTH - 2, BUTTON_HEIGHT - 2, 4, chColors[nButton]);
        spriteBackground.setTextColor(TFT_BLACK, chColors[nButton]);
        spriteBackground.drawChar(chSymbols[(nRow * 4) + nColumn], 
                                  nX + (BUTTON_WIDTH / 2) - (spriteBackground.textWidth(String(chSymbols[nButton])) / 2), 
                                  nY + (BUTTON_HEIGHT / 2) - (spriteBackground.fontHeight() / 2));
      }
    }
  }
}

//
// DrawReadout.
// Draw the readout.
// Entry   : nothing
// Returns : nothing.
//

void  DrawReadout()
{
  // Draw the readout buffer.

  spriteBackground.fillRect(0, (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4)) - (displayTFT.fontHeight() + 4), DISPLAY_WIDTH, displayTFT.fontHeight(), TFT_BLACK);
  spriteBackground.setTextColor(TFT_WHITE, TFT_BLACK);
  spriteBackground.drawString(chReadout, 
                              DISPLAY_WIDTH - displayTFT.textWidth(chReadout), 
                              (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4)) - (displayTFT.fontHeight() + 4));
}

//
// Math
// Perform math.
// Entry   : nothing
// Returns : nothing
//

void  Math()
{
  // Do the math.
  
  switch(chOperator)
  {
    case  '+':
    {
      // Addition.
      
      sprintf(chReadout, "%lf", dOperands[0] + dOperands[1]);
      #if DEBUG
      Serial.println(String(dOperands[0]) + " " + String(chOperator) + " " + String(dOperands[1]) +  " = " + String(chReadout));
      #endif
      dOperands[0] = atof(chReadout);
   }
    break;
    
    case  '-':
    {
      // Subtraction.
      
      sprintf(chReadout, "%lf", dOperands[0] - dOperands[1]);
      #if DEBUG
      Serial.println(String(dOperands[0]) + " " + String(chOperator) + " " + String(dOperands[1]) +  " = " + String(chReadout));
      #endif
      dOperands[0] = atof(chReadout);
    }
    break;
    
    case  '*':
    {
      // Multiplication.
      
      sprintf(chReadout, "%lf", dOperands[0] * dOperands[1]);
      #if DEBUG
      Serial.println(String(dOperands[0]) + " " + String(chOperator) + " " + String(dOperands[1]) +  " = " + String(chReadout));
      #endif
      dOperands[0] = atof(chReadout);
    }
    break;
    
    case  '/':
    {
      // Division.
      
      if(atof(chReadout) == 0.0)
      {
        // Division by zero.

        sprintf(chReadout, "    ERROR!    ");
        bError = true;
      }
      else
      {
        // Division.
        
        sprintf(chReadout, "%lf", dOperands[0] / dOperands[1]);
        #if DEBUG
        Serial.println(String(dOperands[0]) + " " + String(chOperator) + " " + String(dOperands[1]) +  " = " + String(chReadout));
        #endif
        dOperands[0] = atof(chReadout);
      }
    }
    break;
  }

  // Remove trailing zeros from the readout.

  if(bError == false)
  {
    while(chReadout[strlen(chReadout) - 1] == '0')
        chReadout[strlen(chReadout) - 1] = 0;
  }

  // Check for overflow.

  if((bError == false) && 
     (strlen(chReadout) >= READOUT_CHARACTER_WIDTH))
  {
    // Overflow.
    
    sprintf(chReadout, "  OVERFLOW !  ");
    bError = true;
  }
  
  #if DEBUG
  Serial.println(String(dOperands[0]) + " " + String(chOperator) + " " + String(dOperands[1]) +  "\r\n");
  #endif
}

//
// ProcessKeyPresses
// Process key presses.
// Entry   : the pressed key symbol
// Returns : nothing
//

void  ProcessKeyPresses(char chSymbol)
{
  // Determine which key was pressed.
  
  switch(chSymbol)
  {
    case 'C':
    {
      // Clear key, reset error flag.

      bError = false;

      // Reset math.
      
      memset(dOperands, 0, sizeof(dOperands));
      nOperands = 0;
      chOperator = 0;
      bEqual = bOperator = false;

      // Clear greg on a poll.
      
      spriteBackground.fillRect(0, 0, DISPLAY_WIDTH, (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4)) - (displayTFT.fontHeight() + 6), TFT_WHITE);

      // Reset the readout.

      nReadoutPosition = 0;
      sprintf(chReadout, "0.");

      // Set the clear flag.

      bClearKeyPressed = true;
    }
    break;

    case '.':
    {
      // Decimal point.

      if((strlen(chReadout) < READOUT_CHARACTER_WIDTH) &&
         (bError == false))
      {
        // Have room and no errors, check for an operator.

        if(chOperator)
        {
          // Have an operator, set the display to "0." and prepare to receive digits to the right of the decimal point.
          
          sprintf(chReadout, "0.");
          nReadoutPosition = strlen(chReadout);
          chReadout[nReadoutPosition] = 0;
        }
        else
        {
          // Do not have an operator, prepare to receive digits to the right of the decimal point.
          
          nReadoutPosition = strlen(chReadout);
          chReadout[nReadoutPosition] = 0;
        }

        // Draw the readout.
        
        DrawReadout();
      }
    }
    break;

    case '0':
    {
      // For leading zero suppression, process digit 0 separately from digits 1 through 9.

      if((strlen(chReadout) < READOUT_CHARACTER_WIDTH) &&
         ((chReadout[0] != '0') || (strcspn(chReadout, ".") < nReadoutPosition)) &&
         (bError == false))
      {
        // Have room, not a leading zero, and no errors so add the 0.
      
        chReadout[nReadoutPosition ++] = chSymbol;
        chReadout[nReadoutPosition] = 0;
       
        // Add a decimal point if the user is entering digits to the left of the decimal point.
        
        if(strcspn(chReadout, ".") >= nReadoutPosition)
        {
          chReadout[nReadoutPosition] = '.';
          chReadout[nReadoutPosition + 1] = 0;
        }

        // Draw the readout.

        DrawReadout();
      }
    }
    break;
    
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    {
      // Process digits 1 through 9.

      if((strlen(chReadout) < READOUT_CHARACTER_WIDTH) &&
         (bError == false))
      {
        // Have room and no errors, add the digit.
        
        chReadout[nReadoutPosition ++] = chSymbol;
        chReadout[nReadoutPosition] = 0;
      
        // Add a decimal point after the digit if the user is entering digits to the left of the decimal point.
  
        if(strcspn(chReadout, ".") >= nReadoutPosition)
        {
          chReadout[nReadoutPosition] = '.';
          chReadout[nReadoutPosition + 1] = 0;
        }

        // Draw the readout.
        
        DrawReadout();
      }
    }
    break;

    case '+':
    case '-':
    case '*':
    case '/':
    {
      // Process operators.

      #if DEBUG
      Serial.println(String(chOperator) + ", " + String(nOperands));
      #endif
      
      if(bError == false)
      {
        // No errors, check for operator and operands.
        
        if((chOperator == 0) &&
           (chOperator == 0) &&
           (atof(chReadout) != 0.0))
        {
          // Have no operand, no operators, but have a displayed input, set the operand and displayed input.

          #if DEBUG
          Serial.println("No operand, no operators.");
          #endif
          
          chOperator = chSymbol;
          dOperands[nOperands ++] = atof(chReadout);

          // Reset the readout position.

          nReadoutPosition = 0;
        }
        
        else if((chOperator) &&
                (nOperands == 1))
        {
          // Have one operand and operator, set operand.

          #if DEBUG
          Serial.println("One operand, one operator.");
          #endif
          
          dOperands[nOperands ++] = atof(chReadout);
          
          // Do the math.

          Math();

          // Draw the readout.
          
          DrawReadout();

          // Set the operator.
          
          chOperator = chSymbol;
          
          // Reset the readout position.
          
          nReadoutPosition = 0;

          // Operator pressed.
          
          bOperator = true;
        }
        
        else if((chOperator) &&
                (nOperands == 2))
        {
          #if DEBUG
          Serial.println("Two operands, one operator.");
          #endif

          // Transition from equal to operator.
          
          if(bEqual)
          {
            Serial.println("bEqual == true.");
            
            // Only one operand.

            dOperands[1] = 0;
            nOperands = 1;

            // Set the operator.

            chOperator = chSymbol;
            
            // Reset equal pressed flag.

            bEqual = false;
          }
          else
          {
            #if DEBUG
            Serial.println("bEqual == false.");
            #endif

            // Have two operands and operator, set the operand.
  
            dOperands[1] = atof(chReadout);
            
            // Do the math.
  
            Math();
  
            // Draw the readout.
            
            DrawReadout();
  
            // Set the operator.
            
            chOperator = chSymbol;
            
            // Reset the readout position.
            
            nReadoutPosition = 0;

            // Operator pressed.
            
            bOperator = true;
          }
        }
      }
    }
    break;

    case '=':
    {
      // Process equal.
      
      if(bError == false)
      {
        // No errors, check for operator and operands.
        
        if((chOperator) &&
           (nOperands == 1))
        {
          // Have operator and operand, set the operand.

          dOperands[1] = atof(chReadout);

          // Now have two operands.
          
          nOperands = 2;
          
          // Do the math.

          Math();

          // Draw the readout.
          
          DrawReadout();

          // Reset the readout position.
          
          nReadoutPosition = 0;
        }

        else if((chOperator) &&
               (nOperands == 2))
        {
          // Have operator and two operands, transition from operator to equal.
          
          if(bOperator)
          {
            // Set the operand.

            dOperands[1] = atof(chReadout);
            
            // Reset the operator pressed flag.

            bOperator = false;
          }
          
          // Do the math.

          Math();

          // Draw the readout.
          
          DrawReadout();

          // Reset the readout position.
          
          nReadoutPosition = 0;
        }
    
        // Set the equal flag.
    
        bEqual = true;
      }
    }
    break;
  }
}

//
// ScanKeypad
// Scan the keypad.
// Entry   : nothing
// Returns : button symbol on keypad press, 'C' on non keypad press.
//

char  ScanKeypad()
{
  // Compute the touched kepad button, or 'C' if the touch occured off the keypad.
  
  int         nButtonTouched = (tpTouch.y > (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4))) 
                                ? (tpTouch.x / BUTTON_WIDTH) + ((tpTouch.y - (DISPLAY_HEIGHT - (BUTTON_HEIGHT * 4))) / BUTTON_HEIGHT * 4)
                                : 'C';


  // Return the button symbol for the keypad button that was touched, or "C" if the keypad was not touched.
  
  return (nButtonTouched == 'C') ? nButtonTouched : chSymbols[nButtonTouched];
}

//
// UpdateTouchStatus
// Update touch staus and associated global variables.
// Entry   : none
// Returns : true if event occured
//

bool  UpdateTouchStatus()
{
  // Assume release event.

  bTouchpadPressed = false;
  
  // Update touch status.

  if(tlTouch.read())
  {
    // Touch detected.

    if(nTouchN1 == 0)
    {
      // Touch just pressed.
      
      bTouchpadPressed = true;
      nTouchN1 = 1;
      tpTouch = tlTouch.getPoint(0);

      // Display compensation.
      
      nTouchX = (tpTouch.y * DISPLAY_WIDTH) / TOUCH_MAX_X;
      nTouchX = (nTouchX < 0) ? 0 : (nTouchX > (DISPLAY_WIDTH - 1)) ? DISPLAY_WIDTH - 1 : nTouchX;
      nTouchY = ((170 - tpTouch.x) * DISPLAY_HEIGHT) / TOUCH_MAX_Y;
      nTouchY = (nTouchY < 0) ? 0 : (nTouchY > DISPLAY_HEIGHT - 1) ? DISPLAY_HEIGHT - 1 : nTouchY;

      // Event of interest occured.

      return(true);
    }
    else
    {
      // Track continuous touch for touch release.
      
      tpTouch = tlTouch.getPoint(0);

      // Display compensation.
      
      nTouchX = (tpTouch.y * DISPLAY_WIDTH) / TOUCH_MAX_X;
      nTouchX = (nTouchX < 0) ? 0 : (nTouchX > (DISPLAY_WIDTH - 1)) ? DISPLAY_WIDTH - 1 : nTouchX;
      nTouchY = ((170 - tpTouch.x) * DISPLAY_HEIGHT) / TOUCH_MAX_Y;
      nTouchY = (nTouchY < 0) ? 0 : (nTouchY > DISPLAY_HEIGHT - 1) ? DISPLAY_HEIGHT - 1 : nTouchY;

      // Event of interest did not occur.
      
      return(false);
    }
  }
  else
  {
    if(nTouchN1)
    {
      // Touch just released.

      nTouchN1 = 0;
      bClearKeyPressed = false;
      
      // Event of interest occured.
      
      return(true);
    }
  }
}
